#include "common.h"
#include "httpclient.h"

namespace limax {
	namespace http {
		bool HttpConnection::strless::operator()(const std::string& l, const std::string& r) const
		{
			return http::util::stringcompare(l.c_str(), r.c_str()) < 0;
		}
		HttpConnection::HttpConnection() {}
		HttpConnection::~HttpConnection() {}
		HttpConnectionFactory::HttpConnectionFactory() {}
		HttpConnectionFactory::~HttpConnectionFactory() {}
		std::shared_ptr<HttpConnectionFactory> HttpConnectionFactory::create(int contimeout, int rwtimeout)
		{
			return std::make_shared<http::ConnectionFactoryImpl>(contimeout, rwtimeout);
		}
		std::shared_ptr<HttpConnectionFactory> HttpConnectionFactory::create(const std::string& s, short p, int contimeout, int rwtimeout)
		{
			return std::make_shared<http::ProxyConnectionFactoryImpl>(s, p, contimeout, rwtimeout);
		}
		namespace helper {
			static const char HEX_CHARS[] = "0123456789abcdef";
			inline static std::string toFileNameString(const std::string& s)
			{
				std::ostringstream oss;
				for (auto c : s)
				{
					if ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'))
						oss << c;
					else
						oss << '-' << HEX_CHARS[c >> 4 & 0xf] << HEX_CHARS[c & 0xf];
				}
				return oss.str();
			}
			inline static std::string toFullFileNameString(const std::string& cacheDir, const std::string& s)
			{
				if (cacheDir.back() == '/' || cacheDir.back() == '\\')
					return cacheDir + toFileNameString(s);
				else
					return cacheDir + "/" + toFileNameString(s);
			}
		} // namespace helper {
		std::string httpDownload(const std::string& url, int timeout, size_t maxsize, const std::string& cacheDir, bool staleEnable)
		{
			std::string etag;
			std::string content;
			if (!cacheDir.empty())
			{
				try
				{
					FileUnmarshalStreamSource	fuss(helper::toFullFileNameString(cacheDir, url));
					UnmarshalStream us(fuss);
					us >> etag >> content;
				}
				catch (MarshalException&)
				{
					etag.clear();
					content.clear();
				}
			}
			auto connection = HttpConnectionFactory::create(timeout, timeout)->create(url);
			if (!etag.empty())
				connection->setRequestProperty("If-None-Match", etag);
			if (!connection->doRequest())
			{
				if (Trace::isWarnEnabled())
				{
					std::ostringstream oss;
					oss << "httpDownload doRequest return false : url = " << url;
					Trace::warn(oss.str());
				}
				return staleEnable ? content : "";
			}
			const auto responseCode = connection->getResponseCode();
			if (responseCode == HttpConnection::HTTP_NOT_MODIFIED)
				return content;
			if (responseCode != HttpConnection::HTTP_OK)
				return staleEnable ? content : "";
			content.clear();
			auto getdata = connection->getContent();
			while (true)
			{
				auto pd = getdata();
				if (0 == pd.size())
					break;
				if (pd.size() + content.size() > maxsize)
				{
					if (Trace::isWarnEnabled())
					{
						std::ostringstream oss;
						oss << "httpDownload  size > maxsize : url = " << url << " maxsize = " << maxsize;
						Trace::warn(oss.str());
					}
					return "";
				}
				content.append((const char*)pd.begin(), pd.size());
			}
			if (!cacheDir.empty())
			{
				etag = connection->getRequestProperty("ETag");
				FileMarshalStreamSource	fmss(helper::toFullFileNameString(cacheDir, url));
				MarshalStream us(fmss);
				us << etag << content;
			}
			return content;
		}

		struct WebSocketImpl : public WebSocket, public TcpClient::Listener {
			TcpClient* tcpclient;
			std::string uri;
			enum ReadyState { CONNECTING, OPEN, CLOSING, CLOSED };
			std::atomic<ReadyState> readyState;
			Octets buffer;
			OctetsUnmarshalStreamSource datasrc;
			std::function<void(std::shared_ptr<WebSocket>)> __onopen;
			std::function<void(std::shared_ptr<WebSocket>, WebSocket::CloseStatus)> __onclose;
			std::function<void(std::shared_ptr<WebSocket>, const std::string&)> __onmessage;
			std::weak_ptr<WebSocketImpl> self;
			WebSocketImpl(const std::string& ip, short port, const std::string& _uri, std::function<void(std::shared_ptr<WebSocket>)> _onopen, std::function<void(std::shared_ptr<WebSocket>, WebSocket::CloseStatus)> _onclose, std::function<void(std::shared_ptr<WebSocket>, const std::string&)> _onmessage) : tcpclient(nullptr), uri(_uri), readyState(CONNECTING), datasrc(buffer), __onopen(_onopen), __onclose(_onclose), __onmessage(_onmessage)
			{
			}
			~WebSocketImpl()
			{
				if (tcpclient)
					tcpclient->destroy();
			}
			virtual void onCreate(TcpClient *tcpclient) override
			{
				this->tcpclient = tcpclient;
			}
			virtual void onOpen(const IPEndPoint& local, const IPEndPoint& peer) override
			{
				std::string request;
				request.append("GET ").append(uri).append(" HTTP/1.1\r\n");
				request.append("Connection: Upgrade\r\n");
				request.append("Upgrade: WebSocket\r\n");
				request.append("Sec-WebSocket-Version: 13\r\n");
				request.append("Sec-WebSocket-Key: AQIDBAUGBwgJCgsMDQ4PEC==\r\n");
				request.append("Origin: null\r\n");
				request.append("\r\n");
				tcpclient->send(request.c_str(), request.length());
			}
			virtual void onAbort(const IPEndPoint& sa) override
			{
				readyState = CLOSED;
				onclose(WebSocket::CloseStatus::CONNECTION_FAIL);
			}
			virtual void onRecv(const void *data, int32_t size) override
			{
				buffer.insert(buffer.end(), data, size);
				switch (readyState)
				{
				case CONNECTING:{

									size_t size = buffer.size();
									if (size >= 4)
									{
										const char *p = (const char *)buffer.begin();
										if (p[size - 4] == '\r' && p[size - 3] == '\n' &&p[size - 2] == '\r' &&p[size - 1] == '\n')
										{
											readyState = OPEN;
											buffer.clear();
										}
										else
										{
											buffer.erase(0, size - 3);
										}
									}
									break;
				}
				case OPEN:{
							  datasrc.transaction(datasrc.Begin);
							  try
							  {
								  UnmarshalStream is(datasrc);
								  is.pop_byte_8();
								  size_t len = is.pop_byte_8();
								  switch (len)
								  {
								  case 126:
									  len = is.pop_byte_16();
									  break;
								  case 127:
									  len = (size_t)is.pop_byte_64();
								  }
								  char *data = new char[len];
								  is.pop_byte(data, len);
								  onmessage(std::string(data, len));
								  delete[] data;
								  datasrc.transaction(datasrc.Commit);
							  }
							  catch (MarshalException)
							  {
								  datasrc.transaction(datasrc.Rollback);
							  }
							  break;
				}
				}
			}
			virtual void onClose(int status) override
			{
				ReadyState expState = ReadyState::CLOSING;
				if (readyState.compare_exchange_strong(expState, ReadyState::CLOSED, std::memory_order::memory_order_relaxed))
				{
					onclose(WebSocket::CloseStatus::ACTIVE_CLOSED);
				}
				else if (expState == ReadyState::OPEN)
				{
					onclose(WebSocket::CloseStatus::PASSIVE_CLOSED);
				}
			}
			void onopen()
			{
				std::shared_ptr<WebSocketImpl> _this(self);
				runOnUiThread([_this](){_this->__onopen(_this); });
			}
			void onclose(WebSocket::CloseStatus status)
			{
				std::shared_ptr<WebSocketImpl> _this(self);
				runOnUiThread([_this, status](){_this->__onclose(_this, status); });
			}
			void onmessage(const std::string& message)
			{
				std::shared_ptr<WebSocketImpl> _this(self);
				runOnUiThread([_this, message](){_this->__onmessage(_this, message); });
			}
			virtual void send(const std::string& utf8message) override
			{
				OctetsMarshalStream os;
				size_t len = utf8message.length();
				os << (uint8_t)0x81;
				if (len < 126)
					os << (uint8_t)(len | 0x80);
				else if (len < 65536)
					os << (uint8_t)0xfe << (uint16_t)len;
				else
					os << (uint8_t)0xff << (uint64_t)len;
				os << (uint32_t)0;
				os.push_byte(utf8message.c_str(), len);
				tcpclient->send(os.getOctets());
			}
			virtual void close() override
			{
				ReadyState expState = ReadyState::OPEN;
				if (readyState.compare_exchange_strong(expState, ReadyState::CLOSING, std::memory_order::memory_order_relaxed))
				{
					TcpClient* tmp = tcpclient;
					tcpclient = nullptr;
					tmp->destroy();
				}
			}
		};

		std::shared_ptr<WebSocket> createWebSocket(const std::string& ip, short port, const std::string& uri, std::function<void(std::shared_ptr<WebSocket>)> onopen, std::function<void(std::shared_ptr<WebSocket>, WebSocket::CloseStatus)> onclose, std::function<void(std::shared_ptr<WebSocket>, const std::string&)> onmessage)
		{
			std::shared_ptr<WebSocketImpl> ws = std::make_shared<WebSocketImpl>(ip, port, uri, onopen, onclose, onmessage);
			ws->self = std::weak_ptr<WebSocketImpl>(ws);
			TcpClient::createAsync(TcpClient::createWeakListener(ws), ip, port);
			return ws;
		}
	} // namespace http {
} // namespace limax {
